# ** header v3.0
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2021 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2021 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see
# <https://www.gnu.org/licenses/>.
#
# ** end header
#
module Entity

# Creators
export create_entity,
    create_parent, create_property, create_property_entity, create_record, create_recordtype

# getters
export get_id,
    get_role,
    get_name,
    get_description,
    get_datatype,
    get_unit,
    get_value,
    get_version_id,
    get_property,
    get_properties,
    get_parent,
    get_parents,
    get_error,
    get_errors,
    get_warning,
    get_warnings,
    get_info,
    get_infos,
    get_importance,
    get_code

# setters
export append_parent,
    append_parents,
    append_property,
    append_properties,
    remove_parent,
    remove_property,
    set_id,
    set_role,
    set_name,
    set_description,
    set_datatype,
    set_unit,
    set_value,
    set_importance,
    set_local_path,
    set_remote_path

# helper functions
export has_errors, has_warnings

using CaosDB

# TODO(henrik, daniel) How much do we want to help the users when
# setting/getting the values? Should it only be allowed if the
# datatype of the entity/property is set? This might be too strong,
# especially in case of a property which I might want to specify by
# name, add a value, append it to a Record, and then insert it and
# might expect the server to care for the datatype.
#
# Also, how much do we want to support them with typos. Currently, C
# and C++ only understand all-capital strings for datatypes,
# importances, roles. What about user input like "record_type",
# "RecordType", "recordtype"?
"""
Struct containing a pointer to the wrapped cpp entity
object. Meant for internal use; use `CaosDB.Entity.create_entity` to
create an entity.
"""
mutable struct _Entity
    wrapped_entity::Ptr{Cvoid}
    _deletable::Bool

    function _Entity(managed_by_julia::Bool = false)

        entity = new()
        entity._deletable = false

        if managed_by_julia
            function f(t)
                ccall(
                    (:caosdb_entity_delete_entity, CaosDB.library_name),
                    Cint,
                    (Ref{_Entity},),
                    Ref{_Entity}(t),
                )
            end
            finalizer(f, entity)
        end

        return entity
    end
end


"""
Struct containing a pointer to the wrapped cpp parent
object. Meant for internal use; use `CaosDB.Entity.create_parent` to
create an parent.
"""
mutable struct _Parent
    wrapped_parent::Ptr{Cvoid}
    _deletable::Bool

    function _Parent(managed_by_julia::Bool = false)

        parent = new()
        parent._deletable = false

        if managed_by_julia
            function f(t)
                ccall(
                    (:caosdb_entity_delete_parent, CaosDB.library_name),
                    Cint,
                    (Ref{_Parent},),
                    Ref{_Parent}(t),
                )
            end
            finalizer(f, parent)
        end

        return parent
    end
end


"""
Struct containing a pointer to the wrapped cpp property
object. Meant for internal use; use `CaosDB.Entity.create_property` to
create an property.
"""
mutable struct _Property
    wrapped_property::Ptr{Cvoid}
    _deletable::Bool

    function _Property(managed_by_julia::Bool = false)

        property = new()
        property._deletable = false

        if managed_by_julia
            function f(t)
                ccall(
                    (:caosdb_entity_delete_property, CaosDB.library_name),
                    Cint,
                    (Ref{_Property},),
                    Ref{_Property}(t),
                )
            end
            finalizer(f, property)
        end

        return property
    end
end

"""
Struct containing a pointer to the wrapped cpp messsage object. Meant
for internal use only; don't create it by yourself.
"""
mutable struct _Message
    wrapped_message::Ptr{Cvoid}
    _deletable::Bool

    function _Message()
        mess = new()
        mess._deletable = false
        return mess
    end
end

"""
    function create_entity(name::AbstractString = "")

Return a new entity object. If a `name` was provided, set the name of
the entity correspondingly.
"""
function create_entity(name::AbstractString = "")

    entity = Ref{_Entity}(_Entity(true))

    err_code = ccall(
        (:caosdb_entity_create_entity, CaosDB.library_name),
        Cint,
        (Ref{_Entity},),
        entity,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    if name != ""
        set_name(entity, name)
    end

    return entity

end

"""
    function create_recordtype(name::AbstractString = "")

Return a new entity object with role RecordType. If a `name` was
provided, its name is set accordingly.
"""
function create_recordtype(name::AbstractString = "")

    record_type = create_entity(name)

    set_role(record_type, "RECORD_TYPE")

    return record_type

end

"""
     function create_record(name::AbstractString = "")

Return a new entity object with role Record. If a `name` was provided,
its name is set accordingly.
"""
function create_record(name::AbstractString = "")

    record = create_entity(name)

    set_role(record, "RECORD")

    return record

end

"""
    function create_property_entity(;
        name::AbstractString = "",
        datatype::AbstractString = "",
        unit::AbstractString = "",
        is_reference::Bool = false,
        is_list::Bool = false,
    )

Return a new entity object with role Record. If `name`, `datatype`, or
`unit` were provided, its name, datatype (including whether it is a
reference or a list), or unit are set accordingly.
"""
function create_property_entity(;
    name::AbstractString = "",
    datatype::AbstractString = "",
    unit::AbstractString = "",
    is_reference::Bool = false,
    is_list::Bool = false,
)

    property = create_entity(name)

    set_role(property, "PROPERTY")

    if datatype != ""
        set_datatype(property, datatype, is_list = is_list, is_reference = is_reference)
    end

    if unit != ""
        set_unit(property, unit)
    end

    return property
end

"""
    function create_property(;
        name::AbstractString = "",
        id::AbstractString = "",
        value::Union{AbstractString, Number, Bool} = "",
        datatype::AbstractString = "",
        is_reference::Bool = false,
        is_list::Bool = false,
     )

Create a property object that can be appended to another `_Entity`. If
`name`, `id`, or `value` are given, the property's name, id, or value
are set accordingly. The datatype and whether it is a list or a
reference can be specified with `datatype`, `is_reference`, and
`is_list`, respectively.

!!! info

    This is not an `_Entity` that could be inserted or updated by its
    own but a `_Property` that is to be appended to another `_Entity`.
"""
function create_property(;
    name::AbstractString = "",
    id::AbstractString = "",
    value::Union{AbstractString,Number,Bool} = "",
    datatype::AbstractString = "",
    is_reference::Bool = false,
    is_list::Bool = false,
)
    property = Ref{_Property}(_Property(true))

    err_code = ccall(
        (:caosdb_entity_create_property, CaosDB.library_name),
        Cint,
        (Ref{_Property},),
        property,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    if name != ""
        set_name(property, name)
    end

    if id != ""
        set_id(property, id)
    end

    if datatype != ""
        set_datatype(property, datatype, is_reference = is_reference, is_list = is_list)
    end

    if value != ""
        set_value(property, value)
    end

    return property

end

"""
    function create_parent(; name::AbstractString = "", id::AbstractString = "")

Create a parent object that can be appended to another `_Entity`. If
`name`, `id`, or `value` are given, the parent's name, id, or value
are set accordingly.

!!! info

    This is not an `_Entity` that could be inserted or updated by its
    own but a `_Parent` that is to be appended to another `_Entity`.
"""
function create_parent(; name::AbstractString = "", id::AbstractString = "")

    parent = Ref{_Parent}(_Parent(true))

    err_code = ccall(
        (:caosdb_entity_create_parent, CaosDB.library_name),
        Cint,
        (Ref{_Parent},),
        parent,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    if name != ""
        set_name(parent, name)
    end

    if id != ""
        set_id(parent, id)
    end

    return parent
end

"""
    function get_id(entity::Ref{_Entity})

Return the id of the given `entity`
"""
function get_id(entity::Ref{_Entity})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_entity_get_id, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}),
        entity,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_id(entity::Ref{_Parent})

Return the id of the given `parent`
"""
function get_id(parent::Ref{_Parent})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_parent_get_id, CaosDB.library_name),
        Cint,
        (Ref{_Parent}, Ref{Ptr{UInt8}}),
        parent,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_id(entity::Ref{_Property})

Return the id of the given `property`
"""
function get_id(property::Ref{_Property})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_property_get_id, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Ref{Ptr{UInt8}}),
        property,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_role(entity::Ref{_Entity})

Return the role of the given `entity`.
"""
function get_role(entity::Ref{_Entity})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_entity_get_role, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}),
        entity,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_name(entity::Ref{_Entity})

Return the name of the given `entity`
"""
function get_name(entity::Ref{_Entity})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_entity_get_name, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}),
        entity,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_name(entity::Ref{_Parent})

Return the name of the given `parent`
"""
function get_name(parent::Ref{_Parent})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_parent_get_name, CaosDB.library_name),
        Cint,
        (Ref{_Parent}, Ref{Ptr{UInt8}}),
        parent,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_name(entity::Ref{_Property})

Return the name of the given `property`
"""
function get_name(property::Ref{_Property})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_property_get_name, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Ref{Ptr{UInt8}}),
        property,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_description(entity::Ref{_Entity})

Return the description of the given `entity`
"""
function get_description(entity::Ref{_Entity})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_entity_get_description, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}),
        entity,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_description(entity::Ref{_Parent})

Return the description of the given `parent`
"""
function get_description(parent::Ref{_Parent})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_parent_get_description, CaosDB.library_name),
        Cint,
        (Ref{_Parent}, Ref{Ptr{UInt8}}),
        parent,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_description(entity::Ref{_Property})

Return the description of the given `property`
"""
function get_description(property::Ref{_Property})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_property_get_description, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Ref{Ptr{UInt8}}),
        property,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_description(entity::Ref{_Message})

Return the description of the given `message`
"""
function get_description(message::Ref{_Message})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_message_get_description, CaosDB.library_name),
        Cint,
        (Ref{_Message}, Ref{Ptr{UInt8}}),
        message,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

# TODO(henrik, daniel) Is this the best way to return name, is_ref,
# and is_list to the users? Should we split it up?
"""
    function get_datatype(entity::Ref{_Entity})

Return a tuple that contains the name of the datatype of the given
`entity`, whether it is a reference, and whether it is a list.
"""
function get_datatype(entity::Ref{_Entity})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())
    is_ref = Ref{Bool}(false)
    is_list = Ref{Bool}(false)

    err_code = ccall(
        (:caosdb_entity_entity_get_datatype, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}, Ref{Bool}, Ref{Bool}),
        entity,
        out,
        is_ref,
        is_list,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[]), is_ref[], is_list[]
end

"""
    function get_datatype(entity::Ref{_Property})

Return a tuple that contains the name of the datatype of the given
`property`, whether it is a reference, and whether it is a list.
"""
function get_datatype(property::Ref{_Property})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())
    is_ref = Ref{Bool}(false)
    is_list = Ref{Bool}(false)

    err_code = ccall(
        (:caosdb_entity_property_get_datatype, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Ref{Ptr{UInt8}}, Ref{Bool}, Ref{Bool}),
        property,
        out,
        is_ref,
        is_list,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[]), is_ref[], is_list[]
end

"""
    function get_unit(entity::Ref{_Entity})

Return the unit of the given `entity`
"""
function get_unit(entity::Ref{_Entity})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_entity_get_unit, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}),
        entity,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_unit(entity::Ref{_Property})

Return the unit of the given `property`
"""
function get_unit(property::Ref{_Property})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_property_get_unit, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Ref{Ptr{UInt8}}),
        property,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

# TODO(henrik,daniel) Replace ccall within the `get_value` function by
# a check of the datatype and then call the correct function in case
# of scalar values or construct a vector and fill it with all values
# in case of list values.
"""
    function get_value(entity::Ref{_Entity})

Return the value of the given `entity`
"""
function get_value(entity::Ref{_Entity})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_entity_get_value, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}),
        entity,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_value(entity::Ref{_Property})

Return the value of the given `property`
"""
function get_value(property::Ref{_Property})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_property_get_value, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Ref{Ptr{UInt8}}),
        property,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_version_id(entity::Ref{_Entity})

Return the version_id of the given `entity`
"""
function get_version_id(entity::Ref{_Entity})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_entity_get_version_id, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}),
        entity,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_code(message::Ref{_Message})

Return the code of the given `message`.
"""
function get_code(message::Ref{_Message})

    code = Ref{Cint}(0)

    err_code = ccall(
        (:caosdb_entity_message_get_code, CaosDB.library_name),
        Cint,
        (Ref{_Message}, Ref{Cint}),
        message,
        code,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return code

end

"""
    function get_importance(property::Ref{_Property})

Return the importance of the given `property`
"""
function get_importance(property::Ref{_Property})

    out = Ref{Ptr{UInt8}}(Ptr{UInt8}())

    err_code = ccall(
        (:caosdb_entity_property_get_importance, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Ptr{UInt8}}),
        property,
        out,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return unsafe_string(out[])
end

"""
    function get_errors_size(entity::Ref{_Entity})

Return the number of error messages attached to the given `entity`.
"""
function get_errors_size(entity::Ref{_Entity})

    size = Ref{Cint}(0)

    err_code = ccall(
        (:caosdb_entity_entity_get_errors_size, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Cint}),
        entity,
        size,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return size[]

end

"""
    function get_error(entity::Ref{_Entity}, index::Cint)

Return the error message of the given `entity` with the provided
`index`.
"""
function get_error(entity::Ref{_Entity}, index::Cint)

    size = get_errors_size(entity)

    if index > size
        throw(
            DomainError(
                index,
                "You tried to access the error at position $index but the entity only has $size errors.",
            ),
        )
    end

    error = Ref{_Message}(_Message())

    err_code = ccall(
        (:caosdb_entity_entity_get_error, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cint),
        entity,
        index - Cint(1),
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    message_text = get_description(error)

    code = get_code(error)

    return CaosDB.Exceptions.CaosDBMessage(message_text, code)
end

"""
    function get_error(entity::Ref{_Entity}, index::Integer)

Convenience wrapper for `get_error(::Ref{_Entity}, ::Cint)`. Convert
`index` to Int32 and return the error at position `index` of the given
`entity`.
"""
function get_error(entity::Ref{_Entity}, index::Integer)

    return get_error(entity, Cint(index))

end

"""
    function get_warnings_size(entity::Ref{_Entity})

Return the number of warning messages attached to the given `entity`.
"""
function get_warnings_size(entity::Ref{_Entity})

    size = Ref{Cint}(0)

    err_code = ccall(
        (:caosdb_entity_entity_get_warnings_size, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Cint}),
        entity,
        size,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return size[]

end

"""
    function get_warning(entity::Ref{_Entity}, index::Cint)

Return the warning message of the given `entity` with the provided
`index`.
"""
function get_warning(entity::Ref{_Entity}, index::Cint)

    size = get_warnings_size(entity)

    if index > size
        throw(
            DomainError(
                index,
                "You tried to access the warning at position $index but the entity only has $size warnings.",
            ),
        )
    end

    warning = Ref{_Message}(_Message())

    err_code = ccall(
        (:caosdb_entity_entity_get_warning, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cint),
        entity,
        index - Cint(1),
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    message_text = get_description(warning)

    code = get_code(warning)

    return CaosDB.Exceptions.CaosDBMessage(message_text, code)
end

"""
    function get_warning(entity::Ref{_Entity}, index::Integer)

Convenience wrapper for `get_warning(::Ref{_Entity}, ::Cint)`. Convert
`index` to Int32 and return the warning at position `index` of the given
`entity`.
"""
function get_warning(entity::Ref{_Entity}, index::Integer)

    return get_warning(entity, Cint(index))

end

"""
    function get_infos_size(entity::Ref{_Entity})

Return the number of info messages attached to the given `entity`.
"""
function get_infos_size(entity::Ref{_Entity})

    size = Ref{Cint}(0)

    err_code = ccall(
        (:caosdb_entity_entity_get_infos_size, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Cint}),
        entity,
        size,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return size[]

end

"""
    function get_info(entity::Ref{_Entity}, index::Cint)

Return the info message of the given `entity` with the provided
`index`.
"""
function get_info(entity::Ref{_Entity}, index::Cint)

    size = get_infos_size(entity)

    if index > size
        throw(
            DomainError(
                index,
                "You tried to access the info at position $index but the entity only has $size infos.",
            ),
        )
    end

    info = Ref{_Message}(_Message())

    err_code = ccall(
        (:caosdb_entity_entity_get_info, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cint),
        entity,
        index - Cint(1),
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    message_text = get_description(info)

    code = get_code(info)

    return CaosDB.Exceptions.CaosDBMessage(message_text, code)
end

"""
    function get_info(entity::Ref{_Entity}, index::Integer)

Convenience wrapper for `get_info(::Ref{_Entity}, ::Cint)`. Convert
`index` to Int32 and return the info at position `index` of the given
`entity`.
"""
function get_info(entity::Ref{_Entity}, index::Integer)

    return get_info(entity, Cint(index))

end


"""
    function get_errors(entity::Ref{_Entity})

Return a Vector of all error messages attached to the given `entity`.
"""
function get_errors(entity::Ref{_Entity})

    size = get_errors_size(entity)

    errors = [get_error(entity, Cint(ii)) for ii = 1:size]

    return errors
end

"""
    function get_warnings(entity::Ref{_Entity})

Return a Vector of all warning messages attached to the given `entity`.
"""
function get_warnings(entity::Ref{_Entity})

    size = get_warnings_size(entity)

    warnings = [get_warning(entity, Cint(ii)) for ii = 1:size]

    return warnings
end


"""
    function get_warnings(entity::Ref{_Entity})

Return a Vector of all info messages attached to the given `entity`.
"""
function get_infos(entity::Ref{_Entity})

    size = get_infos_size(entity)

    infos = [get_info(entity, Cint(ii)) for ii = 1:size]

    return infos
end

"""
    function get_parents_size(entity::Ref{_Entity})

Return the number of parents attached to the given `entity`.
"""
function get_parents_size(entity::Ref{_Entity})

    size = Ref{Cint}(0)

    err_code = ccall(
        (:caosdb_entity_entity_get_parents_size, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Cint}),
        entity,
        size,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return size[]

end

"""
    function get_parent(entity::Ref{_Entity}, index::Cint)

Return the parent of the given `entity` at position `index`.
"""
function get_parent(entity::Ref{_Entity}, index::Cint)

    size = get_parents_size(entity)

    if index > size
        throw(
            DomainError(
                index,
                "You tried to access the parent at position $index but the entity only has $size parents.",
            ),
        )
    end

    parent = Ref{_Parent}(_Parent())

    err_code = ccall(
        (:caosdb_entity_entity_get_parent, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{_Parent}, Cint),
        entity,
        parent,
        index - Cint(1),
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return parent
end

"""
    function get_parent(entity::Ref{_Entity}, index::Integer)

Convenience wrapper for `get_parent(::Ref{_Entity}, ::Cint)`. Convert
`index` to Int32 and return the parent at position `index` of the given
`entity`.
"""
function get_parent(entity::Ref{_Entity}, index::Integer)

    return get_parent(entity, Cint(index))

end


"""
    function get_parents(entity::Ref{_Entity})

Return the vector of all parents of the given `entity`.
"""
function get_parents(entity::Ref{_Entity})

    size = get_parents_size(entity)

    parents = [get_parent(entity, Cint(ii)) for ii = 1:size]

    return parents
end

"""
    function get_properties_size(entity::Ref{_Entity})

Return the number of properties attached to the given `entity`.
"""
function get_properties_size(entity::Ref{_Entity})

    size = Ref{Cint}(0)

    err_code = ccall(
        (:caosdb_entity_entity_get_properties_size, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{Cint}),
        entity,
        size,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return size[]

end

"""
    function get_property(entity::Ref{_Entity}, index::Cint)

Return the property of the given `entity` at position `index`.
"""
function get_property(entity::Ref{_Entity}, index::Cint)

    size = get_properties_size(entity)

    if index > size
        throw(
            DomainError(
                index,
                "You tried to access the property at position $index but the entity only has $size properties.",
            ),
        )
    end

    property = Ref{_Property}(_Property())

    err_code = ccall(
        (:caosdb_entity_entity_get_property, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{_Property}, Cint),
        entity,
        property,
        index - Cint(1),
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)

    return property
end


"""
    function get_property(entity::Ref{_Entity}, index::Integer)

Convenience wrapper for `get_property(::Ref{_Entity}, ::Cint)`. Convert
`index` to Int32 and return the property at position `index` of the given
`entity`.
"""
function get_property(entity::Ref{_Entity}, index::Integer)

    return get_property(entity, Cint(index))

end


"""
    function get_properties(entity::Ref{_Entity})

Return the vector of all properties of the given `entity`.
"""
function get_properties(entity::Ref{_Entity})

    size = get_properties_size(entity)

    properties = [get_property(entity, Cint(ii)) for ii = 1:size]

    return properties
end

"""
    function set_id(entity::Ref{_Entity}, id::AbstractString)

Throws a `CaosDB.Exceptions.ClientException` since you are not allowed
to set the id of entities.
"""
function set_id(entity::Ref{_Entity}, id::AbstractString)
    throw(
        CaosDB.Exceptions.ClientException(
            "You cannot set the id of an entity object.
If you want to update an entity with this id, you have to retrieve it first.",
        ),
    )
end

"""
    function set_id(parent::Ref{_Parent}, id::AbstractString)

Set the id of the given `parent` object.
"""
function set_id(parent::Ref{_Parent}, id::AbstractString)
    err_code = ccall(
        (:caosdb_entity_parent_set_id, CaosDB.library_name),
        Cint,
        (Ref{_Parent}, Cstring),
        parent,
        id,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function set_id(property::Ref{_Property}, id::AbstractString)

Set the id of the given `property` object.
"""
function set_id(property::Ref{_Property}, id::AbstractString)
    err_code = ccall(
        (:caosdb_entity_property_set_id, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Cstring),
        property,
        id,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function set_role(entity::Ref{_Entity}, role::AbstractString)

Set the role of the given `entity` object.
"""
function set_role(entity::Ref{_Entity}, role::AbstractString)
    err_code = ccall(
        (:caosdb_entity_entity_set_role, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cstring),
        entity,
        role,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end


"""
    function set_name(entity::Ref{_Entity}, name::AbstractString)

Set the name of the given `entity` object.
"""
function set_name(entity::Ref{_Entity}, name::AbstractString)
    err_code = ccall(
        (:caosdb_entity_entity_set_name, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cstring),
        entity,
        name,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end


"""
    function set_name(property::Ref{_Property}, name::AbstractString)

Set the name of the given `property` object.
"""
function set_name(property::Ref{_Property}, name::AbstractString)
    err_code = ccall(
        (:caosdb_entity_property_set_name, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Cstring),
        property,
        name,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end


"""
    function set_name(parent::Ref{_Parent}, name::AbstractString)

Set the name of the given `parent` object.
"""
function set_name(parent::Ref{_Parent}, name::AbstractString)
    err_code = ccall(
        (:caosdb_entity_parent_set_name, CaosDB.library_name),
        Cint,
        (Ref{_Parent}, Cstring),
        parent,
        name,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function set_description(entity::Ref{_Entity}, description::AbstractString)

Set the description of the given `entity` object.
"""
function set_description(entity::Ref{_Entity}, description::AbstractString)
    err_code = ccall(
        (:caosdb_entity_entity_set_description, CaosDB.library_description),
        Cint,
        (Ref{_Entity}, Cstring),
        entity,
        description,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

# TODO(henrik, daniel) Convenience functions for lists and references,
# too?
"""
    function set_datatype(
        entity::Ref{_Entity},
        datatype::AbstractString;
        is_list::Bool = false,
        is_reference::Bool = false,
    )

Set the datatype of the given `entity` object to the given `datatype`
name. Only possible if the role of the entity is "PROPERTY". Specify
whether it is a reference or a list with the optional `is_list` and
`is_reference` keyword arguments. If the datatype is not a reference,
it is checked against the Atomic Datatypes known to CaosDB and an
exception is thrown if it is not known.
"""
function set_datatype(
    entity::Ref{_Entity},
    datatype::AbstractString;
    is_list::Bool = false,
    is_reference::Bool = false,
)

    if get_role(entity) != "PROPERTY"
        throw(
            CaosDB.Exceptions.ClientException(
                "Only entities with role PROPERTY can be assigned a datatype.",
            ),
        )
    end

    err_code = ccall(
        (:caosdb_entity_entity_set_datatype, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cstring, Bool, Bool),
        entity,
        datatype,
        is_reference,
        is_list,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end


"""
    function set_datatype(
        property::Ref{_Property},
        datatype::AbstractString;
        is_list::Bool = false,
        is_reference::Bool = false,
    )

Set the datatype of the given `property` object to the given
`datatype` name. Specify whether it is a reference or a list with the
optional `is_list` and `is_reference` keyword arguments. If the
datatype is not a reference, it is checked against the Atomic
Datatypes known to CaosDB and an exception is thrown if it is not
known.
"""
function set_datatype(
    property::Ref{_Property},
    datatype::AbstractString;
    is_list::Bool = false,
    is_reference::Bool = false,
)
    err_code = ccall(
        (:caosdb_entity_property_set_datatype, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Cstring, Bool, Bool),
        property,
        datatype,
        is_reference,
        is_list,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function set_unit(entity::Ref{_Entity}, unit::AbstractString)

Set the unit of the given `entity` object.
"""
function set_unit(entity::Ref{_Entity}, unit::AbstractString)
    err_code = ccall(
        (:caosdb_entity_entity_set_unit, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cstring),
        entity,
        unit,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end


"""
    function set_unit(property::Ref{_Property}, unit::AbstractString)

Set the unit of the given `property` object.
"""
function set_unit(property::Ref{_Property}, unit::AbstractString)
    err_code = ccall(
        (:caosdb_entity_property_set_unit, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Cstring),
        property,
        unit,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

# TODO(henrik, daniel) There is no check for the correct datatype
# here. Should we add this?
"""
    function set_value(
        entity::Ref{_Entity},
        value::Union{AbstractString,Number,Bool,Vector{T}},
    ) where {T<:Union{AbstractString,Number,Bool}}


Set the value of the given `entity` object. Throw an error if it
doesn't have the correct role. The value must be either string,
Boolean, Integer, Float, or a vector thereof.
"""
function set_value(
    entity::Ref{_Entity},
    value::Union{AbstractString,Number,Bool,Vector{T}},
) where {T<:Union{AbstractString,Number,Bool}}

    if get_role(entity) != "PROPERTY"
        throw(
            CaosDB.Exceptions.ClientException(
                "Only entites with role PROPERTY may be assigned a value.",
            ),
        )
    end

    in_type = typeof(value)
    if in_type <: AbstractString
        err_code = ccall(
            (:caosdb_entity_entity_set_string_value, CaosDB.library_name),
            Cint,
            (Ref{_Entity}, Cstring),
            entity,
            value,
        )
    elseif in_type <: Bool
        err_code = ccall(
            (:caosdb_entity_entity_set_boolean_value, CaosDB.library_name),
            Cint,
            (Ref{_Entity}, Bool),
            entity,
            value,
        )
    elseif in_type <: Integer
        err_code = ccall(
            (:caosdb_entity_entity_set_int_value, CaosDB.library_name),
            Cint,
            (Ref{_Entity}, Clong),
            entity,
            Clong(value),
        )
    elseif in_type <: Number
        err_code = ccall(
            (:caosdb_entity_entity_set_int_value, CaosDB.library_name),
            Cint,
            (Ref{_Entity}, Cdouble),
            entity,
            Cdouble(value),
        )
    else
        # Type is a vector now
        length = Cint(length(value))
        if in_type <: Vector{T} where {T<:AbstractString}
            err_code = ccall(
                (:caosdb_entity_entity_set_string_list_value, CaosDB.library_name),
                Cint,
                (Ref{_Entity}, Ptr{Ptr{Cchar}}, Cint),
                entity,
                value,
                length,
            )
        elseif in_type <: Vector{T} where {T<:Bool}
            err_code = ccall(
                (:caosdb_entity_entity_set_boolean_list_value, CaosDB.library_name),
                Cint,
                (Ref{_Entity}, Ptr{Bool}, Cint),
                entity,
                value,
                length,
            )
        elseif in_type <: Vector{T} where {T<:Integer}
            err_code = ccall(
                (:caosdb_entity_entity_set_int_list_value, CaosDB.library_name),
                Cint,
                (Ref{_Entity}, Ptr{Clong}, Cint),
                entity,
                Vector{Clong}(value),
                length,
            )
        elseif integer <: Vector{T} where {T<:Number}
            err_code = ccall(
                (:caosdb_entity_entity_set_int_list_value, CaosDB.library_name),
                Cint,
                (Ref{_Entity}, Ptr{Cdouble}, Cint),
                entity,
                Vector{Cdouble}(value),
                length,
            )
        end
    end

    CaosDB.Exceptions.evaluate_return_code(err_code)
end


"""
    function set_value(property::Ref{_Property}, value::AbstractString)

Set the value of the given `property` object.
"""
function set_value(
    property::Ref{_Property},
    value::Union{AbstractString,Number,Bool,Vector{T}},
) where {T<:Union{AbstractString,Number,Bool}}

    in_type = typeof(value)
    if in_type <: AbstractString
        err_code = ccall(
            (:caosdb_entity_property_set_string_value, CaosDB.library_name),
            Cint,
            (Ref{_Property}, Cstring),
            property,
            value,
        )
    elseif in_type <: Bool
        err_code = ccall(
            (:caosdb_entity_property_set_boolean_value, CaosDB.library_name),
            Cint,
            (Ref{_Property}, Bool),
            property,
            value,
        )
    elseif in_type <: Integer
        err_code = ccall(
            (:caosdb_entity_property_set_int_value, CaosDB.library_name),
            Cint,
            (Ref{_Property}, Clong),
            property,
            Clong(value),
        )
    elseif in_type <: Number
        err_code = ccall(
            (:caosdb_entity_property_set_int_value, CaosDB.library_name),
            Cint,
            (Ref{_Property}, Cdouble),
            property,
            Cdouble(value),
        )
    else
        # Type is a vector now
        length = Cint(length(value))
        if in_type <: Vector{T} where {T<:AbstractString}
            err_code = ccall(
                (:caosdb_entity_property_set_string_list_value, CaosDB.library_name),
                Cint,
                (Ref{_Property}, Ptr{Ptr{Cchar}}, Cint),
                property,
                value,
                length,
            )
        elseif in_type <: Vector{T} where {T<:Bool}
            err_code = ccall(
                (:caosdb_entity_property_set_boolean_list_value, CaosDB.library_name),
                Cint,
                (Ref{_Property}, Ptr{Bool}, Cint),
                property,
                value,
                length,
            )
        elseif in_type <: Vector{T} where {T<:Integer}
            err_code = ccall(
                (:caosdb_entity_property_set_int_list_value, CaosDB.library_name),
                Cint,
                (Ref{_Property}, Ptr{Clong}, Cint),
                property,
                Vector{Clong}(value),
                length,
            )
        elseif integer <: Vector{T} where {T<:Number}
            err_code = ccall(
                (:caosdb_entity_property_set_int_list_value, CaosDB.library_name),
                Cint,
                (Ref{_Property}, Ptr{Cdouble}, Cint),
                property,
                Vector{Cdouble}(value),
                length,
            )
        end
    end

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function set_importance(property::Ref{_Property}, importance::AbstractString)

Set the importance of the given `property` object.
"""
function set_importance(property::Ref{_Property}, importance::AbstractString)
    err_code = ccall(
        (:caosdb_entity_property_set_importance, CaosDB.library_name),
        Cint,
        (Ref{_Property}, Cstring),
        property,
        importance,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end
"""
    function set_local_path(entity::Ref{_Entity}, path::AbstractString)

Set the local path of the given `entity` object.
"""
function set_local_path(entity::Ref{_Entity}, path::AbstractString)
    err_code = ccall(
        (:caosdb_entity_entity_set_local_path, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cstring),
        entity,
        path,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function set_remote_path(entity::Ref{_Entity}, path::AbstractString)

Set the remote file path of the given `entity` object.
"""
function set_remote_path(entity::Ref{_Entity}, path::AbstractString)
    err_code = ccall(
        (:caosdb_entity_entity_set_remote_path, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cstring),
        entity,
        path,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function append_parent(entity::Ref{_Entity}, parent::Ref{_Parent})

Append the given `parent` to the parents of the given `entity`.
"""
function append_parent(entity::Ref{_Entity}, parent::Ref{_Parent})

    err_code = ccall(
        (:caosdb_entity_entity_append_parent, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{_Parent}),
        entity,
        parent,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function append_parents(entity::Ref{_Entity}, parents::Vector{Base.RefValue{_Parent}})

Append all given `parents` to the parents of the given `entity`.
"""
function append_parents(entity::Ref{_Entity}, parents::Vector{Base.RefValue{_Parent}})

    for parent in parents
        append_parent(entity, parent)
    end
end

"""
    function remove_parent(entity::Ref{_Entity}, index::Cint)

Remove the parent at position `index` from the parents of the given
entity.
"""
function remove_parent(entity::Ref{_Entity}, index::Cint)

    err_code = ccall(
        (:caosdb_entity_entity_remove_parent, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cint),
        entity,
        index - Cint(1),
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function append_property(entity::Ref{_Entity}, property::Ref{_Property})

Append the given `property` to the properties of the given `entity`.
"""
function append_property(entity::Ref{_Entity}, property::Ref{_Property})

    err_code = ccall(
        (:caosdb_entity_entity_append_property, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Ref{_Property}),
        entity,
        property,
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function append_properties(entity::Ref{_Entity}, properties::Vector{Base.RefValue{_Property}})

Append all given `properties` to the properties of the given `entity`.
"""
function append_properties(
    entity::Ref{_Entity},
    properties::Vector{Base.RefValue{_Property}},
)

    for property in properties
        append_property(entity, property)
    end
end

"""
    function remove_property(entity::Ref{_Entity}, index::Cint)

Remove the property at position `index` from the properties of the given
entity.
"""
function remove_property(entity::Ref{_Entity}, index::Cint)

    err_code = ccall(
        (:caosdb_entity_entity_remove_property, CaosDB.library_name),
        Cint,
        (Ref{_Entity}, Cint),
        entity,
        index - Cint(1),
    )

    CaosDB.Exceptions.evaluate_return_code(err_code)
end

"""
    function has_errors(entity::Ref{_Entity})

Return true if the given `entity` has errors, false otherwise.
"""
function has_errors(entity::Ref{_Entity})

    return get_errors_size(entity) > 0

end

"""
    function has_warnings(entity::Ref{_Entity})

Return true if the given `entity` has warnings, false otherwise.
"""
function has_warnings(entity::Ref{_Entity})

    return get_warnings_size(entity) > 0

end

end
